#!/usr/bin/env python3
# coding: utf-8
import collections
import json
import os
import shlex
from pathlib import Path
from traceback import format_exc
from typing import Dict, Iterable, List, Optional
from uuid import uuid4

from cwltool.update import ALLUPDATES
from cwltool.utils import versionstring
from flask import current_app

from genpei.const import RUN_DIR_STRUCTURE, SERVICE_INFO_JSON
from genpei.type import RunRequest, ServiceInfo, State

CWLTOOL_VERSION: str = versionstring().split(" ")[1]
CWL_VERSIONS: List[str] = list(map(str, ALLUPDATES.keys()))


def read_service_info() -> ServiceInfo:
    with SERVICE_INFO_JSON.open(mode="r") as f:
        service_info: ServiceInfo = json.load(f)
    service_info["workflow_engine_versions"]["cwltool"] = CWLTOOL_VERSION
    service_info["workflow_type_versions"]["CWL"]["workflow_type_version"] = \
        CWL_VERSIONS
    service_info["system_state_counts"] = count_state()  # type: ignore

    return service_info


def generate_run_id() -> str:
    return str(uuid4())


def get_run_dir(run_id: str, run_base_dir: Optional[Path] = None) -> Path:
    if run_base_dir is None:
        run_base_dir = current_app.config["RUN_DIR"]

    return run_base_dir.joinpath(run_id[:2]).joinpath(run_id).resolve()


def get_path(run_id: str, file_type: str,
             run_base_dir: Optional[Path] = None) -> Path:
    run_dir: Path = get_run_dir(run_id, run_base_dir)

    return run_dir.joinpath(RUN_DIR_STRUCTURE[file_type])


def read_run_request(run_id: str) -> RunRequest:
    with get_path(run_id, "run_request").open(mode="r") as f:
        run_request: RunRequest = json.load(f)

    return run_request


def write_file(run_id: str, file_type: str, content: str,
               run_base_dir: Optional[Path] = None) -> None:
    file: Path = get_path(run_id, file_type, run_base_dir)
    file.parent.mkdir(parents=True, exist_ok=True)
    with file.open(mode="w") as f:
        f.write(content)


def flatten_wf_engine_params(wf_engine_params: str) -> List[str]:
    wf_engine_params_obj = json.loads(wf_engine_params)
    params: List[str] = []
    for key, val in wf_engine_params_obj.items():
        params.append(key)
        if isinstance(val, list):
            params.append(",".join(val))
        else:
            try:
                params.append(str(val))
            except Exception:
                current_app.logger.debug(format_exc())

    return params


def get_all_run_ids() -> List[str]:
    run_base_dir: Path = current_app.config["RUN_DIR"]
    run_requests: List[Path] = \
        list(run_base_dir.glob(f"**/{RUN_DIR_STRUCTURE['run_request']}"))
    run_ids: List[str] = \
        [run_request.parent.name for run_request in run_requests]

    return run_ids


def get_state(run_id: str) -> State:
    try:
        state_file: Path = get_path(run_id, "state")
        with state_file.open(mode="r") as f:
            str_state: str = \
                [line for line in f.read().splitlines() if line != ""][0]
        state: State = State[str_state]
        return state
    except Exception:
        return State.UNKNOWN


def read_file(run_id: str, file_type: str) -> str:
    try:
        file: Path = get_path(run_id, file_type)
        with file.open(mode="r") as f:
            return f.read()
    except Exception:
        current_app.logger.debug(format_exc())
        return ""


def read_cmd(run_id: str) -> List[str]:
    try:
        cmd_file: Path = get_path(run_id, "cmd")
        with cmd_file.open(mode="r") as f:
            cmd: str = \
                [line for line in f.read().splitlines() if line != ""][0]
            l_cmd: List[str] = shlex.split(cmd)
            return l_cmd
    except Exception:
        current_app.logger.debug(format_exc())
        return []


def get_outputs(run_id: str) -> Dict[str, str]:
    try:
        cmd: List[str] = read_cmd(run_id)
        outdir_ind: int = cmd.index("--outdir") + 1
        outdir_path: Path = Path(cmd[outdir_ind])
        if not outdir_path.is_absolute():
            outdir_path = \
                get_path(run_id, "exe").joinpath(outdir_path).resolve()
        output_files: List[Path] = sorted(list(walk_all_files(outdir_path)))
        outputs: Dict[str, str] = {}
        for output_file in output_files:
            outputs[output_file.name] = str(output_file)
        return outputs
    except Exception:
        current_app.logger.debug(format_exc())
        return {}


def walk_all_files(dir: Path) -> Iterable[Path]:
    for root, dirs, files in os.walk(dir):
        yield Path(root)
        for file in files:
            yield Path(root).joinpath(file)


def count_state() -> Dict[str, int]:
    run_ids: List[str] = get_all_run_ids()
    count: Dict[str, int] = \
        dict(collections.Counter(
            [get_state(run_id).name for run_id in run_ids]))

    return count



#!/usr/bin/env python3
# coding: utf-8
from pathlib import Path
from typing import Dict

DEFAULT_HOST: str = "127.0.0.1"
DEFAULT_PORT: int = 8080
GET_STATUS_CODE: int = 200
POST_STATUS_CODE: int = 200

RUN_DIR_STRUCTURE: Dict[str, str] = {
    "service_info": "service-info.json",
    "run_request": "run_request.json",
    "state": "state.txt",
    "exe": "exe",
    "output": "output",
    "wf_params": "exe/workflow_params.json",
    "start_time": "start_time.txt",
    "end_time": "end_time.txt",
    "exit_code": "exit_code.txt",
    "stdout": "stdout.log",
    "stderr": "stderr.log",
    "pid": "run.pid",
    "cmd": "cmd.txt",
    "sys_error": "sys_error.log"
}

DATE_FORMAT = "%Y-%m-%dT%H:%M:%S"

CANCEL_TIMEOUT = 10

SRC_DIR: Path = Path(__file__).parent.resolve()
SERVICE_INFO_JSON: Path = \
    SRC_DIR.joinpath(RUN_DIR_STRUCTURE["service_info"]).resolve()
DEFAULT_RUN_DIR = SRC_DIR.parent.joinpath("run").resolve()



#!/usr/bin/env python3
# coding: utf-8
import argparse
import os
from argparse import ArgumentParser, Namespace
from pathlib import Path
from traceback import format_exc
from typing import List, Optional

from flask import Flask, Response, current_app, jsonify
from werkzeug.exceptions import HTTPException

from genpei.const import DEFAULT_HOST, DEFAULT_PORT, DEFAULT_RUN_DIR
from genpei.controller import app_bp
from genpei.type import ErrorResponse


def parse_args() -> Namespace:
    parser: ArgumentParser = argparse.ArgumentParser(
        description="Implementation of GA4GH WES OpenAPI specification " +
                    "using cwltool.")

    parser.add_argument(
        "--host",
        nargs=1,
        type=str,
        help="Host address of Flask. (default: 127.0.0.1)"
    )
    parser.add_argument(
        "-p",
        "--port",
        nargs=1,
        type=int,
        help="Port of Flask. (default: 8080)"
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Enable debug mode of Flask."
    )
    parser.add_argument(
        "-r",
        "--run-dir",
        nargs=1,
        type=str,
        help="Specify the run dir. (default: ./run)"
    )

    args: Namespace = parser.parse_args()

    return args


def handle_default_host(host: Optional[List[str]]) -> str:
    if host is None:
        return os.environ.get("GENPEI_HOST", DEFAULT_HOST)

    return host[0]


def handle_default_port(port: Optional[List[str]]) -> int:
    if port is None:
        return int(os.environ.get("GENPEI_PORT", DEFAULT_PORT))

    return int(port[0])


def handle_default_debug(debug: bool) -> bool:
    if debug is False:
        return bool(os.environ.get("GENPEI_DEBUG", False))

    return debug


def handle_default_run_dir(run_dir: Optional[List[str]]) -> Path:
    run_dir_path: Path
    if run_dir is None:
        run_dir_path = Path(os.environ.get("GENPEI_RUN_DIR", DEFAULT_RUN_DIR))
    else:
        run_dir_path = Path(run_dir[0])
    if not run_dir_path.is_absolute():
        run_dir_path = Path.cwd().joinpath(run_dir_path).resolve()

    return run_dir_path


def fix_errorhandler(app: Flask) -> Flask:
    @app.errorhandler(400)
    @app.errorhandler(401)
    @app.errorhandler(403)
    @app.errorhandler(404)
    @app.errorhandler(500)
    def error_handler(error: HTTPException) -> Response:
        res_body: ErrorResponse = {
            "msg": error.description,  # type: ignore
            "status_code": error.code,  # type: ignore
        }
        response: Response = jsonify(res_body)
        response.status_code = error.code  # type: ignore
        return response

    @app.errorhandler(Exception)
    def error_handler_exception(exception: Exception) -> Response:
        current_app.logger.error(exception.args[0])
        current_app.logger.debug(format_exc())
        res_body: ErrorResponse = {
            "msg": "The server encountered an internal error and was " +
                   "unable to complete your request.",
            "status_code": 500,
        }
        response: Response = jsonify(res_body)
        response.status_code = 500
        return response

    return app


def create_app(run_dir: Path) -> Flask:
    app = Flask(__name__)
    app.register_blueprint(app_bp)
    fix_errorhandler(app)
    app.config["RUN_DIR"] = run_dir

    return app


def run(host: str, port: int, debug: bool, run_dir: Path) -> None:
    app: Flask = create_app(run_dir)
    app.run(host=host, port=port, debug=debug)


def main() -> None:
    args: Namespace = parse_args()
    host: str = handle_default_host(args.host)
    port: int = handle_default_port(args.port)
    debug: bool = handle_default_debug(args.debug)
    run_dir: Path = handle_default_run_dir(args.run_dir)

    run(host, port, debug, run_dir)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8
import multiprocessing as mp
import os
import signal
import time
from datetime import datetime
from multiprocessing.context import BaseContext
from multiprocessing.process import BaseProcess
from pathlib import Path
from traceback import print_exc
from typing import Dict, List, Optional

from cwltool.main import run as cwltool
from flask import abort
from flask.globals import current_app
from werkzeug.datastructures import FileStorage
from werkzeug.utils import secure_filename

from genpei.const import CANCEL_TIMEOUT, DATE_FORMAT
from genpei.type import Log, RunLog, RunRequest, ServiceInfo, State
from genpei.util import (flatten_wf_engine_params, get_all_run_ids,
                         get_outputs, get_path, get_state, read_cmd, read_file,
                         read_run_request, read_service_info, write_file)


def validate_run_request(run_request: RunRequest) -> None:
    required_fields: List[str] = ["workflow_params",
                                  "workflow_type",
                                  "workflow_type_version",
                                  "workflow_url"]
    for field in required_fields:
        if field not in run_request:
            abort(400,
                  f"{field} not included in the form data of the request.")


def validate_wf_type(wf_type: str, wf_type_version: str) -> None:
    service_info: ServiceInfo = read_service_info()
    wf_type_versions = service_info["workflow_type_versions"]

    available_wf_types: List[str] = \
        list(map(str, wf_type_versions.keys()))
    if wf_type not in available_wf_types:
        abort(400,
              f"{wf_type}, the workflow_type specified in the " +
              f"request, is not included in {available_wf_types}, " +
              "the available workflow_types.")

    available_wf_versions: List[str] = \
        list(map(str, wf_type_versions[wf_type]["workflow_type_version"]))
    if wf_type_version not in available_wf_versions:
        abort(400,
              f"{wf_type_version}, the workflow_type_version specified in " +
              f"the request, is not included in {available_wf_versions}, " +
              "the available workflow_type_versions.")


def prepare_exe_dir(run_id: str,
                    request_files: Dict[str, FileStorage]) \
        -> None:
    exe_dir: Path = get_path(run_id, "exe")
    exe_dir.mkdir(parents=True, exist_ok=True)
    for file in request_files.values():
        if file.filename != "":
            filename: str = secure_filename(file.filename)
            file.save(exe_dir.joinpath(filename))  # type: ignore


def fork_run(run_id: str, run_request: RunRequest) -> None:
    run_base_dir: Path = current_app.config["RUN_DIR"]
    ctx: BaseContext = mp.get_context("spawn")
    process: BaseProcess = \
        ctx.Process(target=forked_process,
                    args=(run_id, run_request, run_base_dir))
    process.start()  # Non blocking


def forked_process(run_id: str,
                   run_request: RunRequest, run_base_dir: Path) -> None:
    try:
        write_file(run_id, "state", State.INITIALIZING.name, run_base_dir)
        wf_engine_params: List[str] = \
            flatten_wf_engine_params(run_request["workflow_engine_parameters"])
        if "--outdir" not in wf_engine_params:
            wf_engine_params.append("--outdir")
            wf_engine_params.append(
                str(get_path(run_id, "output", run_base_dir)))
        wf_url: str = run_request["workflow_url"]
        wf_params_file: Path = get_path(run_id, "wf_params", run_base_dir)
        all_args: List[str] = [*wf_engine_params, wf_url, str(wf_params_file)]
        write_file(run_id, "cmd",
                   " ".join(["cwltool", *all_args]), run_base_dir)
        os.chdir(get_path(run_id, "exe", run_base_dir))
        ctx: BaseContext = mp.get_context("spawn")
        process: BaseProcess = \
            ctx.Process(target=run_cwltool,
                        args=(run_id, all_args, run_base_dir))
        write_file(run_id, "state", State.RUNNING.name, run_base_dir)
        write_file(run_id, "start_time", datetime.now().strftime(DATE_FORMAT),
                   run_base_dir)
        process.start()
        pid: Optional[int] = process.pid
        if pid is not None:
            write_file(run_id, "pid", str(pid), run_base_dir)
        process.join()  # blocking
        write_file(run_id, "end_time", datetime.now().strftime(DATE_FORMAT),
                   run_base_dir)
        exit_code: Optional[int] = process.exitcode
        if exit_code is not None:
            write_file(run_id, "exit_code", str(exit_code), run_base_dir)
        if exit_code == 0:
            write_file(run_id, "state", State.COMPLETE.name, run_base_dir)
        else:
            write_file(run_id, "state",
                       State.EXECUTOR_ERROR.name, run_base_dir)
    except Exception:
        write_file(run_id, "state", State.SYSTEM_ERROR.name, run_base_dir)
        print_exc(file=get_path(run_id, "sys_error",
                                run_base_dir).open(mode="w"))


def run_cwltool(run_id: str, all_args: List[str], run_base_dir: Path) -> None:
    cwltool(all_args,
            stdout=get_path(run_id, "stdout",
                            run_base_dir).open(mode="w", buffering=1),
            stderr=get_path(run_id, "stderr",
                            run_base_dir).open(mode="w", buffering=1))


def get_run_log(run_id: str) -> RunLog:
    run_log: RunLog = {
        "run_id": run_id,
        "request": read_run_request(run_id),
        "state": get_state(run_id).name,  # type: ignore
        "run_log": get_log(run_id),
        "task_logs": [],
        "outputs": get_outputs(run_id)
    }

    return run_log


def get_log(run_id: str) -> Log:
    str_exit_code: str = read_file(run_id, "exit_code")
    exit_code: int
    if str_exit_code == "":
        exit_code = -999
    else:
        exit_code = int(str_exit_code)

    log: Log = {
        "name": "",
        "cmd": read_cmd(run_id),
        "start_time": read_file(run_id, "start_time"),
        "end_time": read_file(run_id, "end_time"),
        "stdout": read_file(run_id, "stdout"),
        "stderr": read_file(run_id, "stderr"),
        "exit_code": exit_code
    }

    return log


def validate_run_id(run_id: str) -> None:
    all_run_ids: List[str] = get_all_run_ids()
    if run_id not in all_run_ids:
        abort(404,
              f"The run_id {run_id} you requested does not exist, " +
              "please check with GET /runs.")


def cancel_run(run_id: str) -> None:
    state: State = get_state(run_id)
    if state == State.RUNNING:
        write_file(run_id, "state", State.CANCELING.name)
        try:
            pid: int = int(read_file(run_id, "pid"))
            os.kill(pid, signal.SIGTERM)
            count: int = 0
            while count < CANCEL_TIMEOUT:
                time.sleep(1)
                count += 1
                if get_state(run_id) != State.RUNNING:
                    write_file(run_id, "state", State.CANCELED.name)
                    break
        except Exception:
            write_file(run_id, "state", State.UNKNOWN.name)



#!/usr/bin/env python3
# coding: utf-8
from pathlib import Path
from typing import List

from setuptools import setup

BASE_DIR: Path = Path(__file__).parent.resolve()
REQUIREMNTS_TEXT: Path = BASE_DIR.joinpath("requirements.txt")


def read_requirements_txt() -> List[str]:
    with REQUIREMNTS_TEXT.open(mode="r") as f:
        packages: List[str] = \
            [str(line) for line in f.read().splitlines() if line != ""]

    return packages


def main() -> None:
    setup(name="genpei",
          version="1.0.0",
          description="Implementation of GA4GH WES OpenAPI specification " +
                      "using cwltool.",
          author="suecharo",
          author_email="suehiro619@gmail.com",
          url="https://github.com/suecharo/genpei",
          license="Apache2.0",
          python_requires=">=3.6",
          platforms="any",
          include_package_data=True,
          zip_safe=False,
          classifiers=["Programming Language :: Python"],
          packages=["genpei"],
          install_requires=read_requirements_txt(),
          entry_points={
              "console_scripts": [
                  "genpei=genpei.app:main",
              ]
          }
          )


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding: utf-8
import json
from typing import cast

from flask import Blueprint, Response, request
from flask.json import jsonify

from genpei.const import GET_STATUS_CODE, POST_STATUS_CODE
from genpei.run import (cancel_run, fork_run, get_run_log, prepare_exe_dir,
                        validate_run_id, validate_run_request,
                        validate_wf_type)
from genpei.type import (RunId, RunListResponse, RunLog, RunRequest, RunStatus,
                         ServiceInfo, State)
from genpei.util import (generate_run_id, get_all_run_ids, get_state,
                         read_service_info, write_file)

app_bp = Blueprint("genpei", __name__)


@app_bp.route("/service-info", methods=["GET"])
def get_service_info() -> Response:
    """
    May include information related (but not limited to) the workflow
    descriptor formats, versions supported, the WES API versions supported,
    and information about general service availability.
    """
    res_body: ServiceInfo = read_service_info()
    response: Response = jsonify(res_body)
    response.status_code = GET_STATUS_CODE

    return response


@app_bp.route("/runs", methods=["GET"])
def get_runs() -> Response:
    """
    This list should be provided in a stable ordering. (The actual ordering is
    implementation dependent.) When paging through the list, the client should
    not make assumptions about live updates, but should assume the contents of
    the list reflect the workflow list at the moment that the first page is
    requested. To monitor a specific workflow run, use GetRunStatus or
    GetRunLog.
    """
    res_body: RunListResponse = {
        "runs": [],
        "next_page_token": ""
    }
    for run_id in get_all_run_ids():
        res_body["runs"].append({
            "run_id": run_id,
            "state": get_state(run_id).name  # type: ignore
        })
    response: Response = jsonify(res_body)
    response.status_code = GET_STATUS_CODE

    return response


@app_bp.route("/runs", methods=["POST"])
def post_runs() -> Response:
    """
    This endpoint creates a new workflow run and returns a `RunId` to monitor
    its progress.
    """
    run_request: RunRequest = cast(RunRequest, dict(request.form))
    validate_run_request(run_request)
    validate_wf_type(run_request["workflow_type"],
                     run_request["workflow_type_version"])
    run_id: str = generate_run_id()
    write_file(run_id, "run_request", json.dumps(run_request, indent=2))
    write_file(run_id, "wf_params", run_request["workflow_params"])
    prepare_exe_dir(run_id, request.files)
    write_file(run_id, "state", State.QUEUED.name)
    fork_run(run_id, run_request)
    response: Response = jsonify({
        "run_id": run_id
    })
    response.status_code = POST_STATUS_CODE

    return response


@app_bp.route("/runs/<run_id>", methods=["GET"])
def get_runs_id(run_id: str) -> Response:
    """
    This endpoint provides detailed information about a given workflow run.
    The returned result has information about the outputs produced by this
    workflow (if available), a log object which allows the stderr and stdout
    to be retrieved, a log array so stderr/stdout for individual tasks can be
    retrieved, and the overall state of the workflow run (e.g. RUNNING, see
    the State section).
    """
    validate_run_id(run_id)
    res_body: RunLog = get_run_log(run_id)
    response: Response = jsonify(res_body)
    response.status_code = GET_STATUS_CODE

    return response


@app_bp.route("/runs/<run_id>/cancel", methods=["POST"])
def post_runs_id_cancel(run_id: str) -> Response:
    """
    Cancel a running workflow.
    """
    validate_run_id(run_id)
    cancel_run(run_id)
    res_body: RunId = {"run_id": run_id}
    response: Response = jsonify(res_body)
    response.status_code = POST_STATUS_CODE

    return response


@app_bp.route("/runs/<run_id>/status", methods=["GET"])
def get_runs_id_status(run_id: str) -> Response:
    """
    This provides an abbreviated (and likely fast depending on implementation)
    status of the running workflow, returning a simple result with the overall
    state of the workflow run (e.g. RUNNING, see the State section).
    """
    validate_run_id(run_id)
    res_body: RunStatus = {
        "run_id": run_id,
        "state": get_state(run_id).name  # type: ignore
    }
    response: Response = jsonify(res_body)
    response.status_code = GET_STATUS_CODE

    return response
